import re
from typing import Union


def re_escape(fn):
    def arg_escaped(this, *args):
        t = [re.escape(arg) if isinstance(arg, str) else arg for arg in args]
        return fn(this, *t)

    return arg_escaped


class Pattern:
    """A wrapper class for compiled regular expressions"""

    def __init__(self, pattern: str):
        self._pattern = pattern
        self._regex = None

    def __str__(self) -> str:
        return self._pattern

    def __getattr__(self, attr):
        if self._regex is None:
            self._regex = re.compile(self._pattern)
        return getattr(self._regex, attr)

    @re_escape
    def __radd__(self, other: str) -> "Pattern":
        return Pattern(other + self._pattern)

    @re_escape
    def __add__(self, other: Union["Pattern", str]) -> "Pattern":
        return Pattern(f"{self._pattern}{other}")

    @re_escape
    def __or__(self, other: Union["Pattern", str]) -> "Pattern":
        return Pattern(f"(?:{self._pattern}|{other})")

    def __mul__(self, n: int) -> "Pattern":
        return Pattern(f"(?:{self._pattern}{{{n}}})")



import re
from typing import Optional, Union

from .pattern import Pattern


def re_escape(skip_chars=None):
    if callable(skip_chars):
        func = skip_chars
        skip_chars = ""
        return re_escape()(func)

    if skip_chars is None:
        skip_chars = ""

    def decorator(fn):
        def arg_escaped(*args, **kwargs):
            def escape_with_skip(arg):
                if not isinstance(arg, str):
                    return arg

                # Check for literal backslash before escaped chars
                parts = []
                i = 0
                while i < len(arg):
                    if arg[i] == "\\" and i + 1 < len(arg) and arg[i + 1] in skip_chars:
                        parts.append(arg[i : i + 2])  # Keep \char as is
                        i += 2
                    elif arg[i] in skip_chars:
                        parts.append(arg[i])
                        i += 1
                    else:
                        parts.append(re.escape(arg[i]))
                        i += 1
                return "".join(parts)

            escaped_args = [escape_with_skip(arg) for arg in args]
            return fn(*escaped_args, **kwargs)

        return arg_escaped

    return decorator


# Basic constructors
@re_escape
def lit(text: str) -> Pattern:
    """Literal text pattern - always escapes special characters"""
    return Pattern(text)


def regex(pattern: str) -> Pattern:
    """Raw regex pattern - no escaping applied"""
    return Pattern(pattern)


# Common patterns
digit = regex(r"\d")
nondigit = regex(r"\D")
letter = regex(r"\w")
nonletter = regex(r"\W")
word = regex(r"\w+")
space = regex(r"\s")
nonspace = regex(r"\S")
any = regex(".")
tab = regex("\t")
newline = regex("\n")

# Assertions
begin = regex("^")
end = regex("$")

# Boundary matchers
bound = regex(r"\b")
nonbound = regex(r"\B")


@re_escape
def seq(*patterns: Union[Pattern, str]) -> Pattern:
    """Sequence of patterns"""
    return Pattern("".join(map(str, patterns)))


@re_escape
def alt(*patterns: Union[Pattern, str]) -> Pattern:
    """Alternative patterns (OR)"""
    return Pattern(f"(?:{'|'.join(map(str, patterns))})")


def group(p: Pattern, name: Optional[str] = None) -> Pattern:
    """Capture group, optionally named

    Examples:
        group(word)              -> (\\w+)
        group(word, "username")  -> (?P<username>\\w+)
    """
    if name is not None:
        return Pattern(f"(?P<{name}>{p})")
    return Pattern(f"({p})")


# Character classes
@re_escape("-")
def anyof(*cs: str) -> Pattern:
    """Character class [...] - match any single character within the specified set"""
    return Pattern(f"[{''.join(map(str, cs))}]")


@re_escape("-")
def exclude(*cs: str) -> Pattern:
    """Negated character class [^...] - match any single character excluding those in the specified set"""
    return Pattern(f"[^{''.join(map(str, cs))}]")


# Quantifiers
@re_escape
def some(p: Pattern, greedy: bool = True) -> Pattern:
    """One or more times (+)

    Args:
        p: Pattern to repeat
        greedy: If True (default), use greedy matching; if False, use non-greedy matching

    Examples:
        some(word)           -> (?:\\w+)
        some(word, False)    -> (?:\\w+?)
    """
    return Pattern(f"(?:{p})+{'?' if not greedy else ''}")


@re_escape
def maybe(p: Pattern, greedy: bool = True) -> Pattern:
    """Zero or one time (?)

    Args:
        p: Pattern to make optional
        greedy: If True (default), use greedy matching; if False, use non-greedy matching
    """
    return Pattern(f"(?:{p})?{'?' if not greedy else ''}")


@re_escape
def mightsome(p: Pattern, greedy: bool = True) -> Pattern:
    """Zero or more times (*)

    Args:
        p: Pattern to repeat
        greedy: If True (default), use greedy matching; if False, use non-greedy matching
    """
    return Pattern(f"(?:{p})*{'?' if not greedy else ''}")


def times(n: int):
    """Repeat exactly n times"""

    @re_escape
    def repeat(p: Pattern) -> Pattern:
        return Pattern(f"(?:{p}){{{n}}}")

    return repeat


def between(min_n: int, max_n: Optional[int] = None, greedy: bool = True):
    """Repeat between min_n and max_n times

    Args:
        min_n: Minimum number of repetitions
        max_n: Maximum number of repetitions (if None, unlimited)
        greedy: If True (default), use greedy matching; if False, use non-greedy matching
    """
    max_str = str(max_n) if max_n is not None else ""

    @re_escape
    def repeat(p: Pattern) -> Pattern:
        return Pattern(f"(?:{p}){{{min_n},{max_str}}}{'?' if not greedy else ''}")

    return repeat


@re_escape
def ahead(*patterns: Union[Pattern, str]) -> Pattern:
    """Positive lookahead assertion (?=...)

    Examples:
        ahead("foo")  -> (?=foo)
        ahead(word)   -> (?=\\w+)
    """
    return Pattern(f"(?={''.join(map(str, patterns))})")


@re_escape
def not_ahead(*patterns: Union[Pattern, str]) -> Pattern:
    """Negative lookahead assertion (?!...)

    Examples:
        not_ahead("foo")  -> (?!foo)
    """
    return Pattern(f"(?!{''.join(map(str, patterns))})")


@re_escape
def behind(*patterns: Union[Pattern, str]) -> Pattern:
    """Positive lookbehind assertion (?<=...)

    Examples:
        behind("foo")  -> (?<=foo)
    """
    return Pattern(f"(?<={''.join(map(str, patterns))})")


@re_escape
def not_behind(*patterns: Union[Pattern, str]) -> Pattern:
    """Negative lookbehind assertion (?<!...)

    Examples:
        not_behind("foo")  -> (?<!foo)
    """
    return Pattern(f"(?<!{''.join(map(str, patterns))})")


@re_escape
def atomic_group(*patterns: Union[Pattern, str]) -> Pattern:
    """Create an atomic group (?>...)

    Examples:
        atomic_group("foo", word)  -> (?>foo\\w+)
    """
    return Pattern(f"(?>{''.join(map(str, patterns))})")



from typing import Iterator, Match, Optional
from typing import Pattern as RePattern
from typing import Union, overload

class Pattern:
    _pattern: str
    _regex: Optional[RePattern]

    def __init__(self, pattern: str) -> None: ...
    def __str__(self) -> str: ...

    # Re.Pattern properties
    @property
    def pattern(self) -> str: ...
    @property
    def groups(self) -> int: ...
    @property
    def flags(self) -> int: ...
    @property
    def groupindex(self) -> dict[str, int]: ...

    # Regular expression methods
    def match(
        self, string: str, pos: int = ..., endpos: int = ...
    ) -> Optional[Match]: ...
    def fullmatch(
        self, string: str, pos: int = ..., endpos: int = ...
    ) -> Optional[Match]: ...
    def search(
        self, string: str, pos: int = ..., endpos: int = ...
    ) -> Optional[Match]: ...
    def findall(self, string: str, pos: int = ..., endpos: int = ...) -> list[str]: ...
    def finditer(
        self, string: str, pos: int = ..., endpos: int = ...
    ) -> Iterator[Match]: ...
    def split(self, string: str, maxsplit: int = ...) -> list[str]: ...
    def sub(self, repl: Union[str, callable], string: str, count: int = ...) -> str: ...
    def subn(
        self, repl: Union[str, callable], string: str, count: int = ...
    ) -> tuple[str, int]: ...

    # Pattern combination methods
    @overload
    def __add__(self, other: str) -> "Pattern": ...
    @overload
    def __add__(self, other: "Pattern") -> "Pattern": ...
    def __radd__(self, other: str) -> "Pattern": ...
    @overload
    def __or__(self, other: str) -> "Pattern": ...
    @overload
    def __or__(self, other: "Pattern") -> "Pattern": ...
    def __mul__(self, n: int) -> "Pattern": ...
